@@ -1,5 +1,5 @@
 /******************************************************************************
- * Copyright 1994-2008,2009 by Thomas E. Dickey                               *
+ * Copyright 1994-2010,2012 by Thomas E. Dickey                               *
  * All Rights Reserved.                                                       *
  *                                                                            *
  * Permission to use, copy, modify, and distribute this software and its      *
@@ -20,7 +20,7 @@
  ******************************************************************************/
 
 #ifndef	NO_IDENT
-static const char *Id = "$Id: diffstat.c,v 1.47 2009/03/29 18:04:40 tom Exp $";
+static const char *Id = "$Id: diffstat.c,v 1.55 2012/01/03 09:44:24 tom Exp $";
 #endif
 
 /*
@@ -28,6 +28,40 @@ static const char *Id = "$Id: diffstat.c,v 1.47 2009/03/29 18:04:40 tom Exp $";
  * Author:	T.E.Dickey
  * Created:	02 Feb 1992
  * Modified:
+ *		03 Jan 2012, Correct case for "xz" suffix in is_compressed()
+ *			     (patch from Frederic Culot in FreeBSD ports).  Add
+ *			     "-R" option.  Improve dequoting of filenames in
+ *			     headers.
+ *		10 Oct 2010, correct display of new files when -S/-D options
+ *			     are used.  Remove the temporary directory on
+ *			     error, introduced in 1.48+ (patch by Solar
+ *			     Designer).
+ *		19 Jul 2010, add missing "break" statement which left "-c"
+ *			     option falling-through into "-C".
+ *		16 Jul 2010, configure "xz" path explicitly, in case lzcat
+ *			     does not support xz format.  Add "-s" (summary)
+ *			     and "-C" (color) options.
+ *		15 Jul 2010, fix strict gcc warnings, e.g., using const.
+ *		10 Jan 2010, improve a case where filenames have embedded blanks
+ *			     (patch by Reinier Post).
+ *		07 Nov 2009, correct suffix-check for ".xz" files as
+ *			     command-line parameters rather than as piped
+ *			     input (report by Moritz Barsnick).
+ *		06 Oct 2009, fixes to build/run with MSYS or MinGW.  use
+ *			     $TMPDIR for path of temporary file used in
+ *			     decompression.  correct else-condition for
+ *			     detecting compression type (patch by Zach Hirsch).
+ *		31 Aug 2009, improve lzma support, add support for xz (patch by
+ *			     Eric Blake).  Add special case for no-newline
+ *			     message from some diff's (Ubuntu #269895).
+ *			     Improve configure check for getopt().
+ *		11 Aug 2009, Add logic to check standard input, decompress if
+ *			     possible.  Add -N option, to truncate long names.
+ *			     Add pack/pcat as a compression type.
+ *			     Add lzma/lzcat as a compression type.
+ *			     Allow overriding program paths with environment.
+ *		10 Aug 2009, modify to work with Perforce-style diffs (patch
+ *			     by Ed Schouten).
  *		29 Mar 2009, modify to work with patch ".rej" files, which have
  *			     no filename header (use the name of the ".rej"
  *			     file if it is available).
@@ -140,7 +174,7 @@ static const char *Id = "$Id: diffstat.c,v 1.47 2009/03/29 18:04:40 tom Exp $";
 #include <config.h>
 #endif
 
-#ifdef WIN32
+#if defined(WIN32) && !defined(HAVE_CONFIG_H)
 #define HAVE_STDLIB_H
 #define HAVE_STRING_H
 #define HAVE_MALLOC_H
@@ -180,16 +214,16 @@ extern int isatty(int);
 #undef HAVE_TSEARCH
 #endif
 
-#ifdef HAVE_FGETC_UNLOCKED
-#define MY_FGETC fgetc_unlocked
+#ifdef HAVE_GETC_UNLOCKED
+#define MY_GETC getc_unlocked
 #else
-#define MY_FGETC fgetc
+#define MY_GETC getc
 #endif
 
 #ifdef HAVE_GETOPT_H
 #include <getopt.h>
-#else
-extern int getopt();
+#elif !defined(HAVE_GETOPT_HEADER)
+extern int getopt(int, char *const *, const char *);
 extern char *optarg;
 extern int optind;
 #endif
@@ -202,21 +236,51 @@ extern int optind;
 #define EXIT_FAILURE 1
 #endif
 
+#ifndef BZCAT_PATH
+#define BZCAT_PATH ""
+#endif
+
+#ifndef BZIP2_PATH
+#define BZIP2_PATH ""
+#endif
+
 #ifndef COMPRESS_PATH
 #define COMPRESS_PATH ""
 #endif
 
 #ifndef GZIP_PATH
-#define GZIP_PATH     ""
+#define GZIP_PATH ""
 #endif
 
-#ifndef BZIP_PATH
-#define BZIP_PATH     ""
+#ifndef LZCAT_PATH
+#define LZCAT_PATH ""
+#endif
+
+#ifndef PCAT_PATH
+#define PCAT_PATH ""
+#endif
+
+#ifndef UNCOMPRESS_PATH
+#define UNCOMPRESS_PATH ""
+#endif
+
+#ifndef XZ_PATH
+#define XZ_PATH ""
+#endif
+
+#ifndef ZCAT_PATH
+#define ZCAT_PATH ""
 #endif
 
 /******************************************************************************/
 
-#ifdef WIN32
+#if defined(__MINGW32__) || defined(WIN32)
+#define MKDIR(name,mode) mkdir(name)
+#else
+#define MKDIR(name,mode) mkdir(name,mode)
+#endif
+
+#if defined(WIN32) && !defined(__MINGW32__)
 #define PATHSEP '\\'
 #else
 #define PATHSEP '/'
@@ -274,6 +338,7 @@ typedef enum {
 typedef struct _data {
     struct _data *link;
     char *name;			/* the filename */
+    int copy;			/* true if filename is const-literal */
     int base;			/* beginning of name if -p option used */
     Comment cmt;
     int pending;
@@ -282,19 +347,36 @@ typedef struct _data {
     long count[MARKS];		/* counts for the file */
 } DATA;
 
+typedef enum {
+    dcNone = 0,
+    dcBzip,
+    dcCompress,
+    dcGzip,
+    dcLzma,
+    dcPack,
+    dcXz,
+    dcEmpty
+} Decompress;
+
 static const char marks[MARKS + 1] = "+-!=";
+static const int colors[MARKS + 1] =
+{2, 1, 6, 4};
 
 static DATA *all_data;
-static char *comment_opt = "";
+static const char *comment_opt = "";
 static char *path_opt = 0;
 static int format_opt = FMT_NORMAL;
 static int max_width;		/* the specified width-limit */
 static int merge_names = 1;	/* true if we merge similar filenames */
 static int merge_opt = 0;	/* true if we merge ins/del as modified */
-static int name_wide;		/* the amount reserved for filenames */
+static int min_name_wide;	/* minimum amount reserved for filenames */
+static int max_name_wide;	/* maximum amount reserved for filenames */
 static int names_only;		/* true if we list filenames only */
 static int num_marks = 3;	/* 3 or 4, according to "-P" option */
+static int reverse_opt;		/* true if results are reversed */
+static int show_colors;		/* true if showing SGR colors */
 static int show_progress;	/* if not writing to tty, show progress */
+static int summary_only = 0;	/* true if only summary line is shown */
 static int path_dest;		/* true if path_opt is destination (patched) */
 static int plot_width;		/* the amount left over for histogram */
 static int prefix_opt = -1;	/* if positive, controls stripping of PATHSEP */
@@ -338,7 +420,7 @@ is_dir(const char *name)
 {
     struct stat sb;
     return (stat(name, &sb) == 0 &&
-	    S_ISDIR(sb.st_mode));
+	    (sb.st_mode & S_IFMT) == S_IFDIR);
 }
 
 static void
@@ -365,33 +447,34 @@ compare_data(const void *a, const void *b)
 }
 
 static void
-init_data(DATA * data, char *name, int base)
+init_data(DATA * data, const char *name, int copy, int base)
 {
     memset(data, 0, sizeof(*data));
-    data->name = name;
+    data->name = (char *) name;
+    data->copy = copy;
     data->base = base;
     data->cmt = Normal;
 }
 
 static DATA *
-new_data(char *name, int base)
+new_data(const char *name, int base)
 {
     DATA *r = (DATA *) xmalloc(sizeof(DATA));
 
-    init_data(r, new_string(name), base);
+    init_data(r, new_string(name), 0, base);
 
     return r;
 }
 
 #ifdef HAVE_TSEARCH
 static DATA *
-add_tsearch_data(char *name, int base)
+add_tsearch_data(const char *name, int base)
 {
     DATA find;
     DATA *result;
     void *pp;
 
-    init_data(&find, name, base);
+    init_data(&find, name, 1, base);
     if ((pp = tfind(&find, &sorted_data, compare_data)) != 0) {
 	result = *(DATA **) pp;
 	return result;
@@ -406,7 +489,7 @@ add_tsearch_data(char *name, int base)
 #endif
 
 static DATA *
-find_data(char *name)
+find_data(const char *name)
 {
     DATA *p, *q, *r;
     DATA find;
@@ -422,7 +505,7 @@ find_data(char *name)
 	    char *s = strchr(name + base, PATHSEP);
 	    if (s == 0 || *++s == EOS)
 		break;
-	    base = s - name;
+	    base = (int) (s - name);
 	}
 	TRACE(("** base set to %d\n", base));
     }
@@ -441,7 +524,7 @@ find_data(char *name)
     } else
 #endif
     {
-	init_data(&find, name, base);
+	init_data(&find, name, 1, base);
 	for (p = all_data, q = 0; p != 0; q = p, p = p->link) {
 	    int cmp = compare_data(p, &find);
 	    if (merge_names && (cmp == 0))
@@ -483,7 +566,8 @@ delink(DATA * data)
 		q->link = p->link;
 	    else
 		all_data = p->link;
-	    free(p->name);
+	    if (!p->copy)
+		free(p->name);
 	    free(p);
 	    return 1;
 	}
@@ -491,9 +575,13 @@ delink(DATA * data)
     return 0;
 }
 
-/* like strncmp, but without the 3rd argument */
-static int
-match(const char *s, const char *p)
+/*
+ * Compare string 's' against a constant, returning either a pointer just
+ * past the matched part of 's' if it matches exactly, or null if a mismatch
+ * was found.
+ */
+static char *
+match(char *s, const char *p)
 {
     int ok = 0;
 
@@ -509,7 +597,7 @@ match(const char *s, const char *p)
 	    break;
 	}
     }
-    return ok;
+    return ok ? s : 0;
 }
 
 static int
@@ -537,8 +625,8 @@ edit_range(const char *s)
  */
 static int
 decode_default(char *s,
-	       int *first, int *first_size,
-	       int *second, int *second_size)
+	       long *first, long *first_size,
+	       long *second, long *second_size)
 {
     int rc = 0;
     char *next;
@@ -567,7 +655,7 @@ decode_default(char *s,
 			*second_size = strtol(s, &next, 10) + 1 - *second;
 		    }
 		}
-		if (next != 0 && next != s && *next == '\0')
+		if (next != 0 && next != s && *next == EOS)
 		    rc = 1;
 		break;
 	    }
@@ -634,93 +722,174 @@ is_leaf(const char *theLeaf, const char *path)
     return 0;
 }
 
+static char *
+trim_datapath(DATA ** datap, size_t length, int *localp)
+{
+    char *target = (*datap)->name;
+
+#ifdef HAVE_TSEARCH
+    /*
+     * If we are using tsearch(), make a local copy of the data
+     * so we can trim it without interfering with tsearch's
+     * notion of the ordering of data.  That will create some
+     * spurious empty data, so we add the changed() macro in a
+     * few places to skip over those.
+     */
+    if (use_tsearch) {
+	char *trim = new_string(target);
+	trim[length] = EOS;
+	*datap = add_tsearch_data(trim, (*datap)->base);
+	target = (*datap)->name;
+	free(trim);
+	*localp = 1;
+    } else
+#endif
+	target[length] = EOS;
+
+    return target;
+}
+
+/*
+ * The 'data' parameter points to the first of two markers, while
+ * 'path' is the pathname from the second marker.
+ *
+ * On the first call for
+ * a given file, the 'data' parameter stores no differences.
+ */
 static char *
 do_merging(DATA * data, char *path, int *freed)
 {
-    TRACE(("** do_merging(%s,%s) diffs:%d\n", data->name, path, HadDiffs(data)));
+    char *target = reverse_opt ? path : data->name;
+    char *source = reverse_opt ? data->name : path;
+    char *result = source;
+
+    TRACE(("** do_merging(\"%s\",\"%s\") diffs:%d\n",
+	   data->name, path, HadDiffs(data)));
 
     *freed = 0;
-    if (!HadDiffs(data)) {	/* the data was the first of 2 markers */
-	if (is_leaf(data->name, path)) {
-	    TRACE(("** is_leaf: %s vs %s\n", data->name, path));
-	    *freed = delink(data);
-	} else if (can_be_merged(data->name)
-		   && can_be_merged(path)) {
-	    size_t len1 = strlen(data->name);
-	    size_t len2 = strlen(path);
-	    unsigned n;
+    if (!HadDiffs(data)) {
+
+	if (is_leaf(target, source)) {
+	    TRACE(("** is_leaf: \"%s\" vs \"%s\"\n", target, source));
+	    if (reverse_opt) {
+		TRACE((".. no action @%d\n", __LINE__));
+	    } else {
+		*freed = delink(data);
+	    }
+	} else if (can_be_merged(target)
+		   && can_be_merged(source)) {
+	    size_t len1 = strlen(target);
+	    size_t len2 = strlen(source);
+	    size_t n;
 	    int matched = 0;
 	    int diff = 0;
 	    int local = 0;
 
 	    /*
-	     * Strip suffixes such as ".orig", ".bak", "~", etc.  The current
-	     * data for the merge is in 'path'.  The previous data (in
-	     * 'data->name') may also be a temporary filename (which would not
-	     * be merged since it has no apparent relationship to the current).
+	     * If the source/target differ only by some suffix, e.g., ".orig"
+	     * or ".bak", strip that off.  The target may may also be a
+	     * temporary filename (which would not be merged since it has no
+	     * apparent relationship to the current).
 	     */
 	    if (len1 > len2) {
-		if (!strncmp(data->name, path, len2)) {
-		    TRACE(("** trimming data '%s' to '%.*s'\n",
-			   data->name, (int) len2, data->name));
-		    len1 = len2;
-#ifdef HAVE_TSEARCH
-		    /*
-		     * If we are using tsearch(), make a local copy of the data
-		     * so we can trim it without interfering with tsearch's
-		     * notion of the ordering of data.  That will create some
-		     * spurious empty data, so we add the changed() macro in a
-		     * few places to skip over those.
-		     */
-		    if (use_tsearch) {
-			char *trim = new_string(data->name);
-			trim[len1] = EOS;
-			data = add_tsearch_data(trim, data->base);
-			free(trim);
-			local = 1;
-		    } else
-#endif
-			data->name[len1] = EOS;
+		if (!strncmp(target, source, len2)) {
+		    TRACE(("** trimming data \"%s\" to \"%.*s\"\n",
+			   target, (int) len2, target));
+		    if (reverse_opt) {
+			TRACE((".. no action @%d\n", __LINE__));
+		    } else {
+			target = trim_datapath(&data, len1 = len2, &local);
+		    }
 		}
 	    } else if (len1 < len2) {
-		if (!strncmp(data->name, path, len1)) {
-		    TRACE(("** trimming path '%s' to '%.*s'\n",
-			   path, (int) len1, path));
-		    path[len2 = len1] = EOS;
+		if (!strncmp(target, source, len1)) {
+		    TRACE(("** trimming source \"%s\" to \"%.*s\"\n",
+			   source, (int) len1, source));
+		    if (reverse_opt) {
+			TRACE((".. no action @%d\n", __LINE__));
+		    } else {
+			source[len2 = len1] = EOS;
+		    }
 		}
 	    }
 
-	    for (n = 1; n <= len1 && n <= len2; n++) {
-		if (data->name[len1 - n] != path[len2 - n]) {
-		    diff = n;
-		    break;
+	    /*
+	     * If there was no "-p" option, look for the best match by
+	     * stripping prefixes from both source/target strings.
+	     */
+	    if (prefix_opt < 0) {
+		/*
+		 * Now (whether or not we trimmed a suffix), scan back from the
+		 * end of source/target strings to find if they happen to share
+		 * a common ending, e.g., a/b/c versus d/b/c.  If the strings
+		 * are not identical, then 'diff' will be set, but if they have
+		 * a common ending then 'matched' will be set.
+		 */
+		for (n = 1; n <= len1 && n <= len2; n++) {
+		    if (target[len1 - n] != source[len2 - n]) {
+			diff = (int) n;
+			break;
+		    }
+		    if (source[len2 - n] == PATHSEP) {
+			matched = (int) n;
+		    }
+		}
+
+		TRACE(("** merge @%d, prefix_opt=%d matched=%d diff=%d\n",
+		       __LINE__, prefix_opt, matched, diff));
+		if (matched != 0 && diff) {
+		    if (reverse_opt) {
+			TRACE((".. no action @%d\n", __LINE__));
+		    } else {
+			result = source + ((int) len2 - matched + 1);
+		    }
 		}
-		if (path[len2 - n] == PATHSEP)
-		    matched = n;
 	    }
 
-	    if (prefix_opt < 0
-		&& matched != 0
-		&& diff)
-		path += len2 - matched + 1;
-
-	    if (!local)
-		*freed = delink(data);
-	    TRACE(("** merge @%d, prefix_opt=%d matched=%d diff=%d\n",
-		   __LINE__, prefix_opt, matched, diff));
-	} else if (!can_be_merged(path)) {
-	    TRACE(("** do not merge, retain @%d\n", __LINE__));
-	    /* must not merge, retain existing name */
-	    path = data->name;
+	    if (!local) {
+		if (reverse_opt) {
+		    TRACE((".. no action @%d\n", __LINE__));
+		} else {
+		    *freed = delink(data);
+		}
+	    }
+	} else if (reverse_opt) {
+	    TRACE((".. no action @%d\n", __LINE__));
+	    if (can_be_merged(source)) {
+		TRACE(("** merge @%d\n", __LINE__));
+	    } else {
+		TRACE(("** do not merge, retain @%d\n", __LINE__));
+		/* must not merge, retain existing name */
+		result = target;
+	    }
 	} else {
+	    if (can_be_merged(source)) {
+		TRACE(("** merge @%d\n", __LINE__));
+		*freed = delink(data);
+	    } else {
+		TRACE(("** do not merge, retain @%d\n", __LINE__));
+		/* must not merge, retain existing name */
+		result = target;
+	    }
+	}
+    } else if (reverse_opt) {
+	TRACE((".. no action @%d\n", __LINE__));
+	if (can_be_merged(source)) {
 	    TRACE(("** merge @%d\n", __LINE__));
-	    *freed = delink(data);
+	    result = target;
+	} else {
+	    TRACE(("** do not merge, retain @%d\n", __LINE__));
+	}
+    } else {
+	if (can_be_merged(source)) {
+	    TRACE(("** merge @%d\n", __LINE__));
+	} else {
+	    TRACE(("** do not merge, retain @%d\n", __LINE__));
+	    result = target;
 	}
-    } else if (!can_be_merged(path)) {
-	path = data->name;
     }
-    TRACE(("** finish do_merging ->%s\n", path));
-    return path;
+    TRACE(("** finish do_merging ->\"%s\"\n", result));
+    return result;
 }
 
 static int
@@ -743,10 +912,31 @@ skip_blanks(char *s)
     return s;
 }
 
+/*
+ * Skip a filename, which may be in quotes, to allow embedded blanks in the
+ * name.
+ */
+static char *
+skip_filename(char *s)
+{
+    if (*s == SQUOTE && s[1] != EOS && strchr(s + 1, SQUOTE)) {
+	++s;
+	while (*s != EOS && (*s != SQUOTE) && isgraph(UC(*s))) {
+	    ++s;
+	}
+	++s;
+    } else {
+	while (*s != EOS && isgraph(UC(*s))) {
+	    ++s;
+	}
+    }
+    return s;
+}
+
 static char *
 skip_options(char *params)
 {
-    while (*params != '\0') {
+    while (*params != EOS) {
 	params = skip_blanks(params);
 	if (*params == '-') {
 	    while (isgraph(UC(*params)))
@@ -755,7 +945,7 @@ skip_options(char *params)
 	    break;
 	}
     }
-    return params;
+    return skip_blanks(params);
 }
 
 /*
@@ -764,7 +954,7 @@ skip_options(char *params)
 static void
 dequote(char *s)
 {
-    int len = strlen(s);
+    size_t len = strlen(s);
     int n;
 
     if (*s == SQUOTE && len > 2 && s[len - 1] == SQUOTE) {
@@ -804,7 +994,7 @@ get_line(char **buffer, size_t *have, FILE *fp)
     int ch;
     size_t used = 0;
 
-    while ((ch = MY_FGETC(fp)) != EOF) {
+    while ((ch = MY_GETC(fp)) != EOF) {
 	if (used + 2 > *have) {
 	    adjust_buffer(buffer, *have *= 2);
 	}
@@ -812,7 +1002,7 @@ get_line(char **buffer, size_t *have, FILE *fp)
 	if (ch == '\n')
 	    break;
     }
-    (*buffer)[used] = '\0';
+    (*buffer)[used] = EOS;
     return (used != 0);
 }
 
@@ -831,16 +1021,49 @@ count_lines(DATA * p)
     int result = -1;
     char *filename = 0;
     char *filetail = data_filename(p);
-    unsigned want = strlen(path_opt) + 2 + strlen(filetail);
+    size_t want = strlen(path_opt) + 2 + strlen(filetail);
     FILE *fp;
     int ch;
 
     if ((filename = malloc(want)) != 0) {
-	sprintf(filename, "%s/%s", path_opt, filetail);
+	int merge = 0;
+
+	if (path_dest) {
+	    size_t path_len = strlen(path_opt);
+	    size_t tail_len;
+	    char *tail_sep = strchr(filetail, PATHSEP);
+
+	    if (tail_sep != 0) {
+		tail_len = (size_t) (tail_sep - filetail);
+		if (tail_len != 0 && tail_len <= path_len) {
+		    if (tail_len < path_len
+			&& path_opt[path_len - tail_len - 1] != PATHSEP) {
+			merge = 0;
+		    } else if (!strncmp(path_opt + path_len - tail_len,
+					filetail,
+					tail_len - 1)) {
+			merge = 1;
+			if (path_len > tail_len) {
+			    sprintf(filename, "%.*s%c%s",
+				    (int) (path_len - tail_len),
+				    path_opt,
+				    PATHSEP,
+				    filetail);
+			} else {
+			    strcpy(filename, filetail);
+			}
+		    }
+		}
+	    }
+	}
+	if (!merge) {
+	    sprintf(filename, "%s%c%s", path_opt, PATHSEP, filetail);
+	}
+
 	TRACE(("count_lines %s\n", filename));
 	if ((fp = fopen(filename, "r")) != 0) {
 	    result = 0;
-	    while ((ch = MY_FGETC(fp)) != EOF) {
+	    while ((ch = MY_GETC(fp)) != EOF) {
 		if (ch == '\n')
 		    ++result;
 	    }
@@ -881,7 +1104,7 @@ finish_chunk(DATA * p)
 	     * are marked as insert/delete.
 	     */
 	    if (p->chunk[cInsert] && p->chunk[cDelete]) {
-		int change;
+		long change;
 		if (p->chunk[cInsert] > p->chunk[cDelete]) {
 		    change = p->chunk[cDelete];
 		} else {
@@ -903,7 +1126,7 @@ finish_chunk(DATA * p)
 #define CASE_TRACE() TRACE(("** handle case for '%c' %d:%s\n", *buffer, ok, that ? that->name : ""))
 
 static void
-do_file(FILE *fp, char *default_name)
+do_file(FILE *fp, const char *default_name)
 {
     static const char *only_stars = "***************";
 
@@ -918,7 +1141,7 @@ do_file(FILE *fp, char *default_name)
     size_t length = 0;
     size_t fixed = 0;
     int ok = HAVE_NOTHING;
-    int marker = -1;
+    int marker;
     int freed = 0;
 
     int unified = 0;
@@ -926,8 +1149,8 @@ do_file(FILE *fp, char *default_name)
     int new_unify = 0;
     int expect_unify = 0;
 
-    int old_dft = 0;
-    int new_dft = 0;
+    long old_dft = 0;
+    long new_dft = 0;
 
     int context = 1;
 
@@ -936,8 +1159,7 @@ do_file(FILE *fp, char *default_name)
     int line_no = 0;
 #endif
 
-    memset(&dummy, 0, sizeof(dummy));
-    dummy.name = "";
+    init_data(&dummy, "", 1, 0);
 
     fixed_buffer(&buffer, fixed = length = BUFSIZ);
     fixed_buffer(&b_fname, length);
@@ -973,7 +1195,7 @@ do_file(FILE *fp, char *default_name)
 	 * "patch -U" can create ".rej" files lacking a filename header,
 	 * in unified format.  Check for those.
 	 */
-	if (line_no == 1 && !strncmp(buffer, "@@", 2)) {
+	if (line_no == 1 && !strncmp(buffer, "@@", (size_t) 2)) {
 	    unified = 2;
 	    that = find_data(default_name);
 	    ok = begin_data(that);
@@ -992,7 +1214,7 @@ do_file(FILE *fp, char *default_name)
 	 * space in column 2.  Because of the missing space, we have to count
 	 * lines to ensure we do not confuse the marker lines.
 	 */
-	marker = -1;
+	marker = 0;
 	if (that != &dummy && !strcmp(buffer, only_stars)) {
 	    finish_chunk(that);
 	    TRACE(("** begin context chunk\n"));
@@ -1008,7 +1230,9 @@ do_file(FILE *fp, char *default_name)
 	    marker = 1;
 	    context = 0;
 	} else if (match(buffer, "*** ")) {
-	    marker = 0;
+	} else if ((old_unify + new_unify) == 0 && match(buffer, "==== ")) {
+	    finish_chunk(that);
+	    unified = 2;
 	} else if ((old_unify + new_unify) == 0 && match(buffer, "--- ")) {
 	    finish_chunk(that);
 	    marker = unified = 1;
@@ -1070,13 +1294,18 @@ do_file(FILE *fp, char *default_name)
 		if (new_unify)
 		    --new_unify;
 		break;
-	    case '\0':
+	    case EOS:
 	    case ' ':
 		if (old_unify)
 		    --old_unify;
 		if (new_unify)
 		    --new_unify;
 		break;
+	    case '\\':
+		if (strstr(buffer, "newline") != 0) {
+		    break;
+		}
+		/* FALLTHRU */
 	    default:
 		TRACE(("?? expected more in chunk\n"));
 		old_unify = new_unify = 0;
@@ -1086,7 +1315,7 @@ do_file(FILE *fp, char *default_name)
 		expect_unify = 2;
 	    }
 	} else {
-	    int old_base, new_base;
+	    long old_base, new_base;
 
 	    unified = 0;
 
@@ -1094,7 +1323,7 @@ do_file(FILE *fp, char *default_name)
 		&& decode_default(buffer,
 				  &old_base, &old_dft,
 				  &new_base, &new_dft)) {
-		TRACE(("DFT %d,%d -> %d,%d\n",
+		TRACE(("DFT %ld,%ld -> %ld,%ld\n",
 		       old_base, old_base + old_dft - 1,
 		       new_base, new_base + new_dft - 1));
 		finish_chunk(that);
@@ -1124,7 +1353,7 @@ do_file(FILE *fp, char *default_name)
 	 */
 	if (marker > 0) {
 	    TRACE(("** have marker=%d, override %s\n", marker, buffer));
-	    (void) strncpy(buffer, "***", 3);
+	    (void) strncpy(buffer, "***", (size_t) 3);
 	}
 
 	/*
@@ -1163,8 +1392,7 @@ do_file(FILE *fp, char *default_name)
 	     */
 	case 'I':
 	    CASE_TRACE();
-	    if (match(buffer, "Index: ")) {
-		s = strrchr(buffer, BLANK);	/* last token is name */
+	    if ((s = match(buffer, "Index: ")) != 0) {
 		s = skip_blanks(s);
 		dequote(s);
 		blip('.');
@@ -1177,10 +1405,14 @@ do_file(FILE *fp, char *default_name)
 
 	case 'd':		/* diff command trace */
 	    CASE_TRACE();
-	    if (match(buffer, "diff ")
-		&& *(s = skip_options(buffer + 5)) != '\0') {
-		s = strrchr(buffer, BLANK);
-		s = skip_blanks(s);
+	    if ((s = match(buffer, "diff ")) != 0
+		&& *(s = skip_options(s)) != EOS) {
+		if (reverse_opt) {
+		    *skip_filename(s) = EOS;
+		} else {
+		    s = skip_filename(s);
+		    s = skip_blanks(s);
+		}
 		dequote(s);
 		blip('.');
 		finish_chunk(that);
@@ -1246,6 +1478,38 @@ do_file(FILE *fp, char *default_name)
 	    }
 	    break;
 
+	case '=':
+	    CASE_TRACE();
+	    if (!(ok & HAVE_PATH)) {
+		int rev;
+
+		if (((sscanf(buffer,
+			     "==== %[^\t #]#%d - %[^\t ]",
+			     b_fname,
+			     &rev,
+			     b_temp1) == 3)
+		     || ((sscanf(buffer,
+				 "==== %[^\t #]#%d (%[^)]) - %[^\t ]",
+				 b_fname,
+				 &rev,
+				 b_temp1,
+				 b_temp2) == 4)))
+		    && !version_num(b_fname)
+		    && !contain_any(b_fname, "*")
+		    && !edit_range(b_fname)) {
+		    TRACE(("** found p4-diff\n"));
+		    prev = that;
+		    finish_chunk(that);
+		    s = do_merging(that, b_fname, &freed);
+		    if (freed)
+			prev = 0;
+		    that = find_data(s);
+		    ok = begin_data(that);
+		    TRACE(("** after merge:%d:%s\n", ok, s));
+		}
+	    }
+	    break;
+
 	case '+':
 	    /* FALL-THRU */
 	case '>':
@@ -1284,14 +1548,24 @@ do_file(FILE *fp, char *default_name)
 	    /* FALL-THRU */
 	case 'b':		/* binary */
 	    CASE_TRACE();
-	    if (match(buffer + 1, "inary files ")) {
+	    if ((s = match(buffer + 1, "inary files ")) != 0) {
+		char *first = skip_blanks(s);
+		/* blindly assume the first filename does not contain " and " */
+		char *at_and = strstr(s, " and ");
 		s = strrchr(buffer, BLANK);
-		if (!strcmp(s, " differ")) {
-		    *s = EOS;
-		    s = strrchr(buffer, BLANK);
+		if ((at_and != NULL) && !strcmp(s, " differ")) {
+		    char *second = skip_blanks(at_and + 5);
+
+		    if (reverse_opt) {
+			*at_and = EOS;
+			s = first;
+		    } else {
+			*s = EOS;
+			s = second;
+		    }
 		    blip('.');
 		    finish_chunk(that);
-		    that = find_data(skip_blanks(s));
+		    that = find_data(s);
 		    that->cmt = Binary;
 		    ok = HAVE_NOTHING;
 		}
@@ -1312,14 +1586,29 @@ do_file(FILE *fp, char *default_name)
     }
 }
 
+static void
+show_color(int color)
+{
+    if (color >= 0)
+	printf("\033[%dm", color + 30);
+    else
+	printf("\033[0;39m");
+}
+
 static long
-plot_bar(long count, int c)
+plot_bar(long count, int c, int color)
 {
     long result = count;
 
+    if (show_colors && result != 0)
+	show_color(color);
+
     while (--count >= 0)
 	(void) putchar(c);
 
+    if (show_colors && result != 0)
+	show_color(-1);
+
     return result;
 }
 
@@ -1329,7 +1618,7 @@ plot_bar(long count, int c)
  * length from getting large.
  */
 static long
-plot_num(long num_value, int c, long *extra)
+plot_num(long num_value, int c, int color, long *extra)
 {
     long product;
     long result = 0;
@@ -1341,7 +1630,7 @@ plot_num(long num_value, int c, long *extra)
 	product = (plot_width * num_value);
 	result = ((product + *extra) / plot_scale);
 	*extra = product - (result * plot_scale) - *extra;
-	plot_bar(result, c);
+	plot_bar(result, c, color);
     }
     return result;
 }
@@ -1381,7 +1670,7 @@ plot_round1(const long num[MARKS])
 	}
     }
     for_each_mark(i) {
-	plot_bar(scaled[i], marks[i]);
+	plot_bar(scaled[i], marks[i], colors[i]);
 	result += scaled[i];
     }
     return result;
@@ -1453,7 +1742,7 @@ plot_round2(const long num[MARKS])
     }
 
     for_each_mark(i) {
-	result += plot_bar(scaled[i], marks[i]);
+	result += plot_bar(scaled[i], marks[i], colors[i]);
     }
 
     return result;
@@ -1484,7 +1773,7 @@ plot_numbers(const DATA * p)
 	switch (round_opt) {
 	default:
 	    for_each_mark(i) {
-		used += plot_num(p->count[i], marks[i], &temp);
+		used += plot_num(p->count[i], marks[i], colors[i], &temp);
 	    }
 	    break;
 	case 1:
@@ -1500,7 +1789,7 @@ plot_numbers(const DATA * p)
 	    if (used > plot_width)
 		printf("%ld", used - plot_width);	/* oops */
 	    else
-		plot_bar(plot_width - used, '.');
+		plot_bar(plot_width - used, '.', 0);
 	}
     }
 }
@@ -1513,8 +1802,11 @@ static void
 show_data(const DATA * p)
 {
     char *name = data_filename(p);
+    int width;
 
-    if (!changed(p)) {
+    if (summary_only) {
+	;
+    } else if (!changed(p)) {
 	;
     } else if (p->cmt == Binary && suppress_binary == 1) {
 	;
@@ -1533,10 +1825,18 @@ show_data(const DATA * p)
     } else if (names_only) {
 	printf("%s\n", name);
     } else {
-	printf("%s %-*.*s|",
-	       comment_opt,
-	       name_wide, name_wide,
-	       name);
+	printf("%s ", comment_opt);
+	if (max_name_wide > 0
+	    && max_name_wide < min_name_wide
+	    && max_name_wide < ((width = (int) strlen(name)))) {
+	    printf("%.*s", max_name_wide, name + (width - max_name_wide));
+	} else {
+	    width = ((max_name_wide > 0 && max_name_wide < min_name_wide)
+		     ? max_name_wide
+		     : min_name_wide);
+	    printf("%-*.*s", width, width, name);
+	}
+	putchar('|');
 	switch (p->cmt) {
 	default:
 	case Normal:
@@ -1584,7 +1884,7 @@ summarize(void)
 
     plot_scale = 0;
     for (p = all_data; p; p = p->link) {
-	int len = strlen(p->name);
+	int len = (int) strlen(p->name);
 
 	if (ignore_data(p))
 	    continue;
@@ -1597,8 +1897,8 @@ summarize(void)
 	 */
 	if (prefix_opt >= 0) {
 	    /* p->base has been computed at node creation */
-	    if (name_wide < (len - p->base))
-		name_wide = (len - p->base);
+	    if (min_name_wide < (len - p->base))
+		min_name_wide = (len - p->base);
 	} else {
 	    /*
 	     * If "-pX" option is not given, strip off any prefix which is
@@ -1630,17 +1930,21 @@ summarize(void)
     for (p = all_data; p; p = p->link) {
 	if (!ignore_data(p)) {
 	    EqlOf(p) = 0;
+	    if (reverse_opt) {
+		int save_ins = InsOf(p);
+		int save_del = DelOf(p);
+		InsOf(p) = save_del;
+		DelOf(p) = save_ins;
+	    }
 	    if (path_opt != 0) {
 		int count = count_lines(p);
 
 		if (count >= 0) {
 		    EqlOf(p) = count - ModOf(p);
-		    if (path_dest) {
+		    if (path_dest != 0) {
 			EqlOf(p) -= InsOf(p);
-			InsOf(p) = 0;
 		    } else {
 			EqlOf(p) -= DelOf(p);
-			DelOf(p) = 0;
 		    }
 		    if (EqlOf(p) < 0)
 			EqlOf(p) = 0;
@@ -1660,12 +1964,12 @@ summarize(void)
     if (prefix_opt < 0) {
 	if (prefix_len < 0)
 	    prefix_len = 0;
-	if ((longest_name - prefix_len) > name_wide)
-	    name_wide = (longest_name - prefix_len);
+	if ((longest_name - prefix_len) > min_name_wide)
+	    min_name_wide = (longest_name - prefix_len);
     }
 
-    name_wide++;		/* make sure it's nonzero */
-    plot_width = (max_width - name_wide - 8);
+    min_name_wide++;		/* make sure it's nonzero */
+    plot_width = (max_width - min_name_wide - 8);
     if (plot_width < 10)
 	plot_width = 10;
 
@@ -1707,26 +2011,155 @@ summarize(void)
 }
 
 #ifdef HAVE_POPEN
-static char *
-is_compressed(const char *name)
+static const char *
+get_program(const char *name, const char *dft)
 {
-    char *verb = 0;
+    const char *result = getenv(name);
+    if (result == 0 || *result == EOS)
+	result = dft;
+    TRACE(("get_program(%s) = %s\n", name, result));
+    return result;
+}
+#define GET_PROGRAM(name) get_program("DIFFSTAT_" #name, name)
+
+static char *
+decompressor(Decompress which, const char *name)
+{
+    const char *verb = 0;
+    const char *opts = "";
     char *result = 0;
     size_t len = strlen(name);
 
-    if (len > 2 && !strcmp(name + len - 2, ".Z")) {
-	verb = COMPRESS_PATH;
-    } else if (len > 3 && !strcmp(name + len - 3, ".gz")) {
-	verb = GZIP_PATH;
-    } else if (len > 4 && !strcmp(name + len - 4, ".bz2")) {
-	verb = BZIP_PATH;
+    switch (which) {
+    case dcBzip:
+	verb = GET_PROGRAM(BZCAT_PATH);
+	if (*verb == '\0') {
+	    verb = GET_PROGRAM(BZIP2_PATH);
+	    opts = "-dc";
+	}
+	break;
+    case dcCompress:
+	verb = GET_PROGRAM(ZCAT_PATH);
+	if (*verb == '\0') {
+	    verb = GET_PROGRAM(UNCOMPRESS_PATH);
+	    opts = "-c";
+	    if (*verb == '\0') {
+		/* not all compress's recognize the options, test this last */
+		verb = GET_PROGRAM(COMPRESS_PATH);
+		opts = "-dc";
+	    }
+	}
+	break;
+    case dcGzip:
+	verb = GET_PROGRAM(GZIP_PATH);
+	opts = "-dc";
+	break;
+    case dcLzma:
+	verb = GET_PROGRAM(LZCAT_PATH);
+	opts = "-dc";
+	break;
+    case dcPack:
+	verb = GET_PROGRAM(PCAT_PATH);
+	break;
+    case dcXz:
+	verb = GET_PROGRAM(XZ_PATH);
+	opts = "-dc";
+	break;
+    case dcEmpty:
+	/* FALLTHRU */
+    case dcNone:
+	break;
     }
     if (verb != 0 && *verb != '\0') {
 	result = (char *) xmalloc(strlen(verb) + 10 + len);
-	sprintf(result, "%s -dc \"%s\"", verb, name);
+	sprintf(result, "%s %s", verb, opts);
+	if (*name != '\0') {
+	    sprintf(result + strlen(result), " \"%s\"", name);
+	}
     }
     return result;
+}
 
+static char *
+is_compressed(const char *name)
+{
+    size_t len = strlen(name);
+    Decompress which;
+
+    if (len > 2 && !strcmp(name + len - 2, ".Z")) {
+	which = dcCompress;
+    } else if (len > 2 && !strcmp(name + len - 2, ".z")) {
+	which = dcPack;
+    } else if (len > 3 && !strcmp(name + len - 3, ".gz")) {
+	which = dcGzip;
+    } else if (len > 4 && !strcmp(name + len - 4, ".bz2")) {
+	which = dcBzip;
+    } else if (len > 5 && !strcmp(name + len - 5, ".lzma")) {
+	which = dcLzma;
+    } else if (len > 3 && !strcmp(name + len - 3, ".xz")) {
+	which = dcXz;
+    } else {
+	which = dcNone;
+    }
+    return decompressor(which, name);
+}
+
+#ifdef HAVE_MKDTEMP
+#define MY_MKDTEMP(path) mkdtemp(path)
+#else
+/*
+ * mktemp is supposedly marked obsolete at the same point that mkdtemp is
+ * introduced.
+ */
+static char *
+my_mkdtemp(char *path)
+{
+    char *result = mktemp(path);
+    if (result != 0) {
+	if (MKDIR(result, 0700) < 0) {
+	    result = 0;
+	}
+    }
+    return path;
+}
+#define MY_MKDTEMP(path) my_mkdtemp(path)
+#endif
+
+static char *
+copy_stdin(char **dirpath)
+{
+    const char *tmp = getenv("TMPDIR");
+    char *result = 0;
+    int ch;
+    FILE *fp;
+
+    if (tmp == 0)
+	tmp = "/tmp/";
+    *dirpath = xmalloc(strlen(tmp) + 12);
+
+    strcpy(*dirpath, tmp);
+    strcat(*dirpath, "/diffXXXXXX");
+    if (MY_MKDTEMP(*dirpath) != 0) {
+	result = xmalloc(strlen(*dirpath) + 10);
+	sprintf(result, "%s/stdin", *dirpath);
+
+	if ((fp = fopen(result, "w")) != 0) {
+	    while ((ch = MY_GETC(stdin)) != EOF) {
+		fputc(ch, fp);
+	    }
+	    fclose(fp);
+	} else {
+	    free(result);
+	    result = 0;
+	    rmdir(*dirpath);	/* Assume that the /stdin file was not created */
+	    free(*dirpath);
+	    *dirpath = 0;
+	}
+    } else {
+	free(*dirpath);
+	*dirpath = 0;
+    }
+    return result;
 }
 #endif
 
@@ -1769,10 +2202,12 @@ usage(FILE *fp)
 	"  -l      list filenames only",
 	"  -m      merge insert/delete data in chunks as modified-lines",
 	"  -n NUM  specify minimum width for the filenames (default: auto)",
+	"  -N NUM  specify maximum width for the filenames (default: auto)",
 	"  -o FILE redirect standard output to FILE",
 	"  -p NUM  specify number of pathname-separators to strip (default: common)",
 	"  -q      suppress the \"0 files changed\" message for empty diffs",
 	"  -r NUM  specify rounding for histogram (0=none, 1=simple, 2=adjusted)",
+	"  -R      assume patch was created with old and new files swapped",
 	"  -S PATH specify location of original files, use for unchanged-count",
 	"  -t      print a table (comma-separated-values) rather than histogram",
 	"  -u      do not sort the input list",
@@ -1814,7 +2249,7 @@ main(int argc, char *argv[])
     max_width = 80;
 
     while ((j = getopt_helper(argc, argv,
-			      "bcdD:e:f:hklmn:o:p:qr:S:tuvVw:", 'h', 'V'))
+			      "bcCdD:e:f:hklmn:N:o:p:qr:RsS:tuvVw:", 'h', 'V'))
 	   != -1) {
 	switch (j) {
 	case 'b':
@@ -1823,11 +2258,17 @@ main(int argc, char *argv[])
 	case 'c':
 	    comment_opt = "#";
 	    break;
+	case 'C':
+	    show_colors = 1;
+	    break;
 #if OPT_TRACE
 	case 'd':
 	    trace_opt = 1;
 	    break;
 #endif
+	case 'D':
+	    set_path_opt(optarg, 1);
+	    break;
 	case 'e':
 	    if (freopen(optarg, "w", stderr) == 0)
 		failed(optarg);
@@ -1848,7 +2289,10 @@ main(int argc, char *argv[])
 	    merge_opt = 1;
 	    break;
 	case 'n':
-	    name_wide = atoi(optarg);
+	    min_name_wide = atoi(optarg);
+	    break;
+	case 'N':
+	    max_name_wide = atoi(optarg);
 	    break;
 	case 'o':
 	    if (freopen(optarg, "w", stdout) == 0)
@@ -1857,15 +2301,18 @@ main(int argc, char *argv[])
 	case 'p':
 	    prefix_opt = atoi(optarg);
 	    break;
-	case 'D':
-	    set_path_opt(optarg, 1);
+	case 'r':
+	    round_opt = atoi(optarg);
+	    break;
+	case 'R':
+	    reverse_opt = 1;
+	    break;
+	case 's':
+	    summary_only = 1;
 	    break;
 	case 'S':
 	    set_path_opt(optarg, 0);
 	    break;
-	case 'r':
-	    round_opt = atoi(optarg);
-	    break;
 	case 't':
 	    table_opt = 1;
 	    break;
@@ -1938,7 +2385,101 @@ main(int argc, char *argv[])
 	    }
 	}
     } else {
-	do_file(stdin, "unknown");
+#ifdef HAVE_POPEN
+	FILE *fp;
+	Decompress which = dcEmpty;
+	char *stdin_dir = 0;
+	char *myfile;
+	char sniff[8];
+	int ch;
+	unsigned got = 0;
+	char *command;
+
+	if ((ch = MY_GETC(stdin)) != EOF) {
+	    which = dcNone;
+	    if (ch == 'B') {	/* perhaps bzip2 (poor magic design...) */
+		sniff[got++] = (char) ch;
+		while (got < 5) {
+		    if ((ch = MY_GETC(stdin)) == EOF)
+			break;
+		    sniff[got++] = (char) ch;
+		}
+		if (got == 5
+		    && !strncmp(sniff, "BZh", (size_t) 3)
+		    && isdigit((unsigned char) sniff[3])
+		    && isdigit((unsigned char) sniff[4])) {
+		    which = dcBzip;
+		}
+	    } else if (ch == ']') {	/* perhaps lzma */
+		sniff[got++] = (char) ch;
+		while (got < 4) {
+		    if ((ch = MY_GETC(stdin)) == EOF)
+			break;
+		    sniff[got++] = (char) ch;
+		}
+		if (got == 4
+		    && !memcmp(sniff, "]\0\0\200", (size_t) 4)) {
+		    which = dcLzma;
+		}
+	    } else if (ch == 0xfd) {	/* perhaps xz */
+		sniff[got++] = (char) ch;
+		while (got < 6) {
+		    if ((ch = MY_GETC(stdin)) == EOF)
+			break;
+		    sniff[got++] = (char) ch;
+		}
+		if (got == 6
+		    && !memcmp(sniff, "\3757zXZ\0", (size_t) 6)) {
+		    which = dcXz;
+		}
+	    } else if (ch == '\037') {	/* perhaps compress, etc. */
+		sniff[got++] = (char) ch;
+		if ((ch = MY_GETC(stdin)) != EOF) {
+		    sniff[got++] = (char) ch;
+		    switch (ch) {
+		    case 0213:
+			which = dcGzip;
+			break;
+		    case 0235:
+			which = dcCompress;
+			break;
+		    case 0036:
+			which = dcPack;
+			break;
+		    }
+		}
+	    } else {
+		sniff[got++] = (char) ch;
+	    }
+	}
+	/*
+	 * The C standard only guarantees one ungetc;
+	 * virtually everyone allows more.
+	 */
+	while (got != 0) {
+	    ungetc(sniff[--got], stdin);
+	}
+	if (which != dcNone
+	    && which != dcEmpty
+	    && (myfile = copy_stdin(&stdin_dir)) != 0) {
+
+	    /* open pipe to decompress temporary file */
+	    command = decompressor(which, myfile);
+	    if ((fp = popen(command, "r")) != 0) {
+		do_file(fp, "stdin");
+		(void) pclose(fp);
+	    }
+	    free(command);
+
+	    unlink(myfile);
+	    free(myfile);
+	    myfile = 0;
+	    rmdir(stdin_dir);
+	    free(stdin_dir);
+	    stdin_dir = 0;
+	} else if (which != dcEmpty)
+#endif
+	    do_file(stdin, "stdin");
     }
     summarize();
 #if defined(NO_LEAKS)
